# -*- mode: python; coding: utf-8 -*-
# Copyright 2020 the AAS WorldWide Telescope project
# Licensed under the MIT License.

"""
Support for loading images from a Djangoplicity database.
"""

__all__ = '''
DjangoplicityImageSource
DjangoplicityCandidateInput
'''.split()

import codecs
from contextlib import contextmanager
from datetime import datetime, timezone
import functools
import html
import json
import numpy as np
import os.path
import requests
import shutil
from urllib.parse import urljoin, quote as urlquote
import yaml

from ..image import ImageLoader
from . import CandidateInput, ImageSource, NotActionableError


class DjangoplicityImageSource(ImageSource):
    """
    An ImageSource that obtains its inputs from a query to a Djangoplicity website.
    """

    _base_url = None
    _channel_name = None
    _search_page_name = None
    _force_insecure_tls = True  # TODO: migrate to False

    @classmethod
    def get_config_key(cls):
        return 'djangoplicity'


    @classmethod
    def deserialize(cls, data):
        inst = cls()
        inst._base_url = data['base_url']
        inst._channel_name = data['channel_name']
        inst._search_page_name = data.get('search_page_name', 'page')
        inst._force_insecure_tls = data.get('force_insecure_tls', True)  # TODO: migrate to false
        return inst

    @contextmanager
    def make_request(self, url, stream=False, none404=False):
        """force_insecure_tls is for noirlab.edu"""

        with requests.get(url, stream=stream, verify=not self._force_insecure_tls) as resp:
            if none404 and resp.status_code == 404:
                yield None
                return

            if not resp.ok:
                raise Exception(f'error fetching url `{url}`: {resp.status_code}')

            if stream:
                # By default, `resp.raw` does not perform content decoding.
                # eso.org gives us gzipped content. The following bit is
                # apparently the preferred workaround. Cf:
                # https://github.com/psf/requests/issues/2155
                #
                # A side effect of the content decoding, however, is that the
                # first read of the stream can return a zero-length string,
                # which causes `readlines` iteration to exit. Callers must be
                # prepared to handle this.
                resp.raw.decode_content = True

            yield resp

    def query_candidates(self):
        page_num = 1

        while True:
            url = self._base_url + f'archive/search/{self._search_page_name}/{page_num}/?type=Observation'
            print(f'requesting {url} ...')

            with self.make_request(url, stream=True, none404=True) as resp:
                if resp is None:
                    break  # got a 404 -- all done

                text_stream = codecs.getreader('utf8')(resp.raw)
                json_lines = []

                # Cf. stream=True in make_request -- skip the zero-length result
                # to prevent readlines iteration from exiting early. This is
                # definitely OK since our `var images` line won't be the first
                # line.
                text_stream.readline()

                for line in text_stream:
                    if not len(json_lines):
                        if 'var images = [' in line:
                            json_lines.append('[')
                    elif '];' in line:
                        json_lines.append(']')
                        break
                    else:
                        json_lines.append(line)

            if not len(json_lines):
                raise Exception(f'error processing url {url}: no "var images" data found')

            # This is really a JS literal, but YAML is compatible enough.
            # JSON does *not* work because the dict keys here aren't quoted.
            data = yaml.safe_load(''.join(json_lines))

            for item in data:
                yield DjangoplicityCandidateInput(item)

            page_num += 1


    def fetch_candidate(self, unique_id, cand_data_stream, cachedir):
        url = self._base_url + urlquote(unique_id) + '/api/json/'

        with self.make_request(url) as resp:
            info = json.loads(resp.content)

        # Find the "fullsize original" image URL

        fullsize_url = None

        for resource in info['Resources']:
            if resource.get('ResourceType') == 'Original':
                fullsize_url = resource['URL']
                break

        if fullsize_url is None:
            raise Exception(f'error processing {unique_id}: can\'t identify \"fullsize original\" image URL')

        ext = fullsize_url.rsplit('.', 1)[-1].lower()
        info['toasty_image_extension'] = ext

        # Validate that there's actually WCS we can use

        if not isinstance(info.get('Spatial.CoordsystemProjection', None), str):
            raise NotActionableError('image does not have full WCS')

        # Download it

        with self.make_request(fullsize_url, stream=True) as resp:
            with open(os.path.join(cachedir, 'image.' + ext), 'wb') as f:
                shutil.copyfileobj(resp.raw, f)

        with open(os.path.join(cachedir, 'metadata.json'), 'wt', encoding='utf8') as f:
            json.dump(info, f, ensure_ascii=False, indent=2)


    def process(self, unique_id, cand_data_stream, cachedir, builder):
        # Set up the metadata.

        with open(os.path.join(cachedir, 'metadata.json'), 'rt', encoding='utf8') as f:
            info = json.load(f)

        img_path = os.path.join(cachedir, 'image.' + info['toasty_image_extension'])
        md = DjangoplicityMetadata(info)

        # Load up the image.

        img = ImageLoader().load_path(img_path)

        # Do the processing.

        builder.tile_base_as_study(img)
        builder.make_thumbnail_from_other(img)

        builder.imgset.set_position_from_wcs(
            md.as_wcs_headers(img.width, img.height),
            img.width,
            img.height,
            place = builder.place,
        )

        builder.set_name(info['Title'])
        builder.imgset.credits_url = info['ReferenceURL']
        builder.imgset.credits = html.escape(info['Credit'])
        builder.place.description = html.escape(info['Description'])

        # Annotation metadata

        pub_dt = datetime.fromisoformat(info['Date'])
        if pub_dt.tzinfo is None:
            pub_dt = pub_dt.replace(tzinfo=timezone.utc)

        amd = {
            'channel': self._channel_name,
            'itemid': unique_id,
            'publishedUTCISO8601': pub_dt.isoformat(),
        }
        builder.place.annotation = json.dumps(amd)

        # Finally, crunch the rest of the pyramid.

        builder.cascade()


class DjangoplicityCandidateInput(CandidateInput):
    """
    A CandidateInput obtained from an AstroPix query.
    """

    def __init__(self, info):
        self._info = info

    def get_unique_id(self):
        return self._info['id']

    def save(self, stream):
        with codecs.getwriter('utf8')(stream) as text_stream:
            json.dump(self._info, text_stream, ensure_ascii=False, indent=2)


class DjangoplicityMetadata(object):
    metadata = None

    def __init__(self, metadata):
        self.metadata = metadata

    def as_wcs_headers(self, width, height):
        headers = {}

        #headers['RADECSYS'] = self.wcs_coordinate_frame  # causes Astropy warnings
        headers['CTYPE1'] = 'RA---' + self.metadata['Spatial.CoordsystemProjection']
        headers['CTYPE2'] = 'DEC--' + self.metadata['Spatial.CoordsystemProjection']
        headers['CRVAL1'] = float(self.metadata['Spatial.ReferenceValue'][0])
        headers['CRVAL2'] = float(self.metadata['Spatial.ReferenceValue'][1])

        # See Calabretta & Greisen (2002; DOI:10.1051/0004-6361:20021327), eqn 186
        crot = np.cos(float(self.metadata['Spatial.Rotation']) * np.pi / 180)
        srot = np.sin(float(self.metadata['Spatial.Rotation']) * np.pi / 180)
        scale0 = float(self.metadata['Spatial.Scale'][0])

        # Seen in noao-02274; guessing how to handle this
        if not self.metadata['Spatial.Scale'][1]:
            scale1 = np.abs(scale0)
        else:
            scale1 = float(self.metadata['Spatial.Scale'][1])

        lam = scale1 / scale0

        headers['PC1_1'] = crot
        headers['PC1_2'] = -lam * srot
        headers['PC2_1'] = srot / lam
        headers['PC2_2'] = crot

        # If we couldn't get the original image, the pixel density used for
        # the WCS parameters may not match the image resolution that we have
        # available. In such cases, we need to remap the pixel-related
        # headers. From the available examples, `wcs_reference_pixel` seems to
        # be 1-based in the same way that `CRPIXn` are. Since in FITS, integer
        # pixel values correspond to the center of each pixel box, a CRPIXn of
        # [0.5, 0.5] (the lower-left corner) should not vary with the image
        # resolution. A CRPIXn of [W + 0.5, H + 0.5] (the upper-right corner)
        # should map to [W' + 0.5, H' + 0.5] (where the primed quantities are
        # the new width and height).

        factor0 = width / float(self.metadata['Spatial.ReferenceDimension'][0])
        factor1 = height / float(self.metadata['Spatial.ReferenceDimension'][1])

        headers['CRPIX1'] = (float(self.metadata['Spatial.ReferencePixel'][0]) - 0.5) * factor0 + 0.5
        headers['CRPIX2'] = (float(self.metadata['Spatial.ReferencePixel'][1]) - 0.5) * factor1 + 0.5
        headers['CDELT1'] = scale0 / factor0
        headers['CDELT2'] = scale1 / factor1

        return headers



# -*- mode: python; coding: utf-8 -*-
# Copyright 2020 the AAS WorldWide Telescope project
# Licensed under the MIT License.

"""
Support for loading images from an AstroPix feed.

TODO: update metadata tomfoolery to match the new things that I've learned. Cf.
the ``wwtdatatool wtml report`` utility and the djangoplicity implementation.

"""

__all__ = '''
AstroPixImageSource
AstroPixCandidateInput
'''.split()

import codecs
from datetime import datetime
import json
import numpy as np
import os.path
import requests
import shutil
from urllib.parse import quote as urlquote

from ..image import ImageLoader
from . import CandidateInput, ImageSource, NotActionableError


EXTENSION_REMAPPING = {
    'jpeg': 'jpg',
}


class AstroPixImageSource(ImageSource):
    """
    An ImageSource that obtains its inputs from a query to the AstroPix service.
    """

    _json_query_url = None

    @classmethod
    def get_config_key(cls):
        return 'astropix'

    @classmethod
    def deserialize(cls, data):
        inst = cls()
        inst._json_query_url = data['json_query_url']
        return inst


    def query_candidates(self):
        with requests.get(self._json_query_url, stream=True) as resp:
            feed_data = json.load(resp.raw)

        for item in feed_data:
            yield AstroPixCandidateInput(item)


    def fetch_candidate(self, unique_id, cand_data_stream, cachedir):
        with codecs.getreader('utf8')(cand_data_stream) as text_stream:
            info = json.load(text_stream)

        lower_id = info['image_id'].lower()
        global_id = info['publisher_id'] + '_' + lower_id

        if info['resource_url'] and len(info['resource_url']):
            source_url = info['resource_url']
        else:
            # Original image not findable. Get the best version available from
            # AstroPix.
            #
            # GROSS: saving this code since I won't be testing this super
            # thoroughly ... but it looks like now the original is being made
            # available consistently?

            ##size = int(info['image_max_boundry'])
            ##if size >= 24000:
            ##    best_astropix_size = 24000
            ##elif size >= 12000:
            ##    best_astropix_size = 12000
            ##elif size >= 6000:
            ##    best_astropix_size = 6000
            ##elif size >= 3000:
            ##    best_astropix_size = 3000
            ##elif size >= 1600:
            ##    best_astropix_size = 1600
            ##elif size > 1024:  # transition point to sizes that are always generated
            ##    best_astropix_size = 1280
            ##elif size > 500:
            ##    best_astropix_size = 1024
            ##elif size > 320:
            ##    best_astropix_size = 500
            ##else:
            ##    best_astropix_size = 320

            source_url = 'http://astropix.ipac.caltech.edu/archive/%s/%s/%s_original.jpg' % (
                urlquote(info['publisher_id']),
                urlquote(lower_id),
                urlquote(global_id),
            )

        # Now ready to download the image.

        ext = source_url.rsplit('.', 1)[-1].lower()
        ext = EXTENSION_REMAPPING.get(ext, ext)

        with requests.get(source_url, stream=True) as resp:
            if not resp.ok:
                raise Exception(f'error downloading {source_url}: {resp.status_code}')

            with open(os.path.join(cachedir, 'image.' + ext), 'wb') as f:
                shutil.copyfileobj(resp.raw, f)


    def process(self, unique_id, cand_data_stream, cachedir, builder):
        # Set up the metadata.

        with codecs.getreader('utf8')(cand_data_stream) as text_stream:
            info = json.load(text_stream)

        if info['resource_url'] and len(info['resource_url']):
            ext = info['resource_url'].rsplit('.', 1)[-1].lower()
            ext = EXTENSION_REMAPPING.get(ext, ext)
        else:
            ext = 'jpg'

        img_path = os.path.join(cachedir, 'image.' + ext)
        md = AstroPixMetadata(info)

        # Load up the image.

        img = ImageLoader().load_path(img_path)

        # Do the processing.

        builder.tile_base_as_study(img)
        builder.make_thumbnail_from_other(img)

        builder.imgset.set_position_from_wcs(
            md.as_wcs_headers(img.width, img.height),
            img.width,
            img.height,
            place = builder.place,
        )

        builder.set_name(info['title'])
        builder.imgset.credits_url = md.get_credit_url()

        builder.cascade()


class AstroPixCandidateInput(CandidateInput):
    """
    A CandidateInput obtained from an AstroPix query.
    """

    def __init__(self, json_dict):
        self._json = json_dict
        self._lower_id = self._json['image_id'].lower()
        self._global_id = self._json['publisher_id'] + '_' + self._lower_id

    def get_unique_id(self):
        return self._global_id.replace('/', '_')

    def save(self, stream):
        # First check that this input is usable. The NRAO feed contains an
        # item like this, and based on my investigations they are just not
        # usable right now because the server APIs don't work. So: skip any
        # like this.
        if '/' in self._json['image_id']:
            raise NotActionableError('AstroPix images with "/" in their IDs aren\'t retrievable')

        # TODO? A few NRAO images have SIN projection. Try to recover them?
        if self._json['wcs_projection'] != 'TAN':
            raise NotActionableError('cannot ingest images in non-TAN projections')

        with codecs.getwriter('utf8')(stream) as text_stream:
            json.dump(self._json, text_stream, ensure_ascii=False, indent=2)


ASTROPIX_FLOAT_ARRAY_KEYS = [
    'wcs_reference_dimension',  # NB: should be ints, but sometimes expressed with decimal points
    'wcs_reference_pixel',
    'wcs_reference_value',
    'wcs_scale',
]

ASTROPIX_FLOAT_SCALAR_KEYS = [
    'wcs_rotation',
]

class AstroPixMetadata(object):
    """
    Metadata derived from AstroPix query results.
    """

    image_id = None
    publisher_id = None
    resource_url = None
    wcs_coordinate_frame = None  # ex: 'ICRS'
    wcs_equinox = None  # ex: 'J2000'
    wcs_projection = None  # ex: 'TAN'
    wcs_reference_dimension = None  # ex: [7416.0, 4320.0]
    wcs_reference_value = None  # ex: [187, 12.3]
    wcs_reference_pixel = None  # ex: [1000.4, 1000.7]; from examples, this seems to be 1-based
    wcs_rotation = None  # ex: -0.07 (deg, presumably)
    wcs_scale = None  # ex: [-6e-7, 6e-7]

    def __init__(self, json_dict):
        # Some massaging for consistency:

        for k in ASTROPIX_FLOAT_ARRAY_KEYS:
            if k in json_dict:
                json_dict[k] = list(map(float, json_dict[k]))

        for k in ASTROPIX_FLOAT_SCALAR_KEYS:
            if k in json_dict:
                json_dict[k] = float(json_dict[k])

        for k, v in json_dict.items():
            setattr(self, k, v)


    def as_wcs_headers(self, width, height):
        headers = {}

        #headers['RADECSYS'] = self.wcs_coordinate_frame  # causes Astropy warnings
        headers['CTYPE1'] = 'RA---' + self.wcs_projection
        headers['CTYPE2'] = 'DEC--' + self.wcs_projection
        headers['CRVAL1'] = self.wcs_reference_value[0]
        headers['CRVAL2'] = self.wcs_reference_value[1]

        # See Calabretta & Greisen (2002; DOI:10.1051/0004-6361:20021327), eqn 186

        crot = np.cos(self.wcs_rotation * np.pi / 180)
        srot = np.sin(self.wcs_rotation * np.pi / 180)
        lam = self.wcs_scale[1] / self.wcs_scale[0]

        headers['PC1_1'] = crot
        headers['PC1_2'] = -lam * srot
        headers['PC2_1'] = srot / lam
        headers['PC2_2'] = crot

        # If we couldn't get the original image, the pixel density used for
        # the WCS parameters may not match the image resolution that we have
        # available. In such cases, we need to remap the pixel-related
        # headers. From the available examples, `wcs_reference_pixel` seems to
        # be 1-based in the same way that `CRPIXn` are. Since in FITS, integer
        # pixel values correspond to the center of each pixel box, a CRPIXn of
        # [0.5, 0.5] (the lower-left corner) should not vary with the image
        # resolution. A CRPIXn of [W + 0.5, H + 0.5] (the upper-right corner)
        # should map to [W' + 0.5, H' + 0.5] (where the primed quantities are
        # the new width and height).

        factor0 = width / self.wcs_reference_dimension[0]
        factor1 = height / self.wcs_reference_dimension[1]

        headers['CRPIX1'] = (self.wcs_reference_pixel[0] - 0.5) * factor0 + 0.5
        headers['CRPIX2'] = (self.wcs_reference_pixel[1] - 0.5) * factor1 + 0.5
        headers['CDELT1'] = self.wcs_scale[0] / factor0
        headers['CDELT2'] = self.wcs_scale[1] / factor1

        return headers


    def get_credit_url(self):
        if self.reference_url:
            return self.reference_url

        return 'http://astropix.ipac.caltech.edu/image/%s/%s' % (
            urlquote(self.publisher_id),
            urlquote(self.image_id),
        )



# -*- mode: python; coding: utf-8 -*-
# Copyright 2019-2020 the AAS WorldWide Telescope project
# Licensed under the MIT License.

from __future__ import absolute_import, division, print_function

import numpy as np
from numpy import testing as nt
import os.path
import pytest
import sys

from . import assert_xml_elements_equal, test_path
from .. import cli
from .. import multi_tan


class TestMultiTan(object):
    WTML = """
<Folder Group="Explorer" Name="TestName">
  <Place Angle="0" DataSetType="Sky" Dec="0.74380165289257" Name="TestName"
         Opacity="100" RA="14.419753086419734" Rotation="0.0"
         ZoomLevel="0.1433599999999144">
    <ForegroundImageSet>
      <ImageSet BandPass="Gamma" BaseDegreesPerTile="0.023893333333319066"
                BaseTileLevel="0" BottomsUp="False" CenterX="216.296296296296"
                CenterY="0.74380165289257" DataSetType="Sky" FileType=".png"
                Name="TestName" OffsetX="4.66666666666388e-05"
                OffsetY="4.66666666666388e-05" Projection="Tan" Rotation="0.0"
                Sparse="True" TileLevels="1" Url="UP{1}/{3}/{3}_{2}.png"
                WidthFactor="2">
        <Credits>CT</Credits>
        <CreditsUrl>CU</CreditsUrl>
        <ThumbnailUrl>TU</ThumbnailUrl>
        <Description>DT</Description>
      </ImageSet>
    </ForegroundImageSet>
  </Place>
</Folder>"""

    # Gross workaround for Python 2.7, where the XML serialization is
    # apparently slightly different from Python >=3 (for Numpy types only, I
    # think?). Fun times. We could avoid this by comparing floats as floats,
    # not text, but then we basically have to learn how to deserialize WTML
    # with the full semantics of the format.

    if sys.version_info.major == 2:
        WTML = (WTML
                .replace('Dec="0.74380165289257"', 'Dec="0.743801652893"')
                .replace('RA="14.419753086419734"', 'RA="14.4197530864"')
                .replace('CenterX="216.296296296296"', 'CenterX="216.296296296"')
                .replace('CenterY="0.74380165289257"', 'CenterY="0.743801652893"')
        )

    # Back to the non-gross stuff.

    def setup_method(self, method):
        from tempfile import mkdtemp
        self.work_dir = mkdtemp()

    def teardown_method(self, method):
        from shutil import rmtree
        rmtree(self.work_dir)

    def work_path(self, *pieces):
        return os.path.join(self.work_dir, *pieces)

    def test_basic(self):
        ds = multi_tan.MultiTanDataSource([test_path('wcs512.fits.gz')])
        ds.compute_global_pixelization()

        from xml.etree import ElementTree as etree
        expected = etree.fromstring(self.WTML)

        folder = ds.create_wtml(
            name = 'TestName',
            url_prefix = 'UP',
            fov_factor = 1.0,
            bandpass = 'Gamma',
            description_text = 'DT',
            credits_text = 'CT',
            credits_url = 'CU',
            thumbnail_url = 'TU',
        )
        assert_xml_elements_equal(folder, expected)

        from ..pyramid import PyramidIO
        pio = PyramidIO(self.work_path('basic'), default_format='npy')
        percentiles = ds.generate_deepest_layer_numpy(pio)

        # These are all hardcoded parameters of this test dataset, derived
        # from a manual processing with checking that the results are correct.
        # Note that to make the file more compressible, I quantized its data,
        # which explains why some of the samples below are identical.

        PERCENTILES = {
            1: 0.098039217,
            99: 0.76862746,
        }

        for pct, expected in PERCENTILES.items():
            nt.assert_almost_equal(percentiles[pct], expected)

        MEAN, TLC, TRC, BLC, BRC = range(5)
        SAMPLES = {
            (0, 0): [0.20828014, 0.20392157, 0.22745098, 0.18431373, 0.20000000],
            (0, 1): [0.22180051, 0.18431373, 0.18823530, 0.16470589, 0.18431373],
            (1, 0): [0.22178716, 0.16470589, 0.18431373, 0.11372549, 0.19607843],
            (1, 1): [0.21140813, 0.18431373, 0.20784314, 0.12549020, 0.14117648],
        }

        for (y, x), expected in SAMPLES.items():
            data = np.load(self.work_path('basic', '1', str(y), '{}_{}.npy'.format(y, x)))
            nt.assert_almost_equal(data.mean(), expected[MEAN])
            nt.assert_almost_equal(data[10,10], expected[TLC])
            nt.assert_almost_equal(data[10,-10], expected[TRC])
            nt.assert_almost_equal(data[-10,10], expected[BLC])
            nt.assert_almost_equal(data[-10,-10], expected[BRC])


    def test_basic_cli(self):
        """Test the CLI interface. We don't go out of our way to validate the
        computations in detail -- that's for the unit tests that probe the
        module directly.

        """
        args = [
            'multi-tan-make-data-tiles',
            '--hdu-index', '0',
            '--outdir', self.work_path('basic_cli'),
            test_path('wcs512.fits.gz')
        ]
        cli.entrypoint(args)



# -*- mode: python; coding: utf-8 -*-
# Copyright 2020 the AAS WorldWide Telescope project
# Licensed under the MIT License.

"""
Generate tiles from a collection of images with associated WCS coordinate
systems.

This module has the following Python package dependencies:

- astropy
- ccdproc (to trim FITS CCD datasets)
- reproject
- shapely (to optimize the projection in reproject)

"""

__all__ = '''
make_lsst_directory_loader_generator
MultiWcsProcessor
'''.split()

import numpy as np
from tqdm import tqdm
import warnings

from .image import Image, ImageMode
from .study import StudyTiling


def make_lsst_directory_loader_generator(dirname, unit=None):
    from astropy.io import fits
    from astropy.nddata import ccddata
    import ccdproc
    from glob import glob
    from os.path import join

    def loader_generator(actually_load_data):
        # Ideally we would just return a shape instead of an empty data array in
        # the case where actually_load_data is false, but we can't use
        # `ccdproc.trim_image()` without having a CCDData in hand, so to get the
        # right shape we're going to need to create a full CCDData anyway.

        for fits_path in glob(join(dirname, '*.fits')):
            # `astropy.nddata.ccddata.fits_ccddata_reader` only opens FITS from
            # filenames, not from an open HDUList, which means that creating
            # multiple CCDDatas from the same FITS file rapidly becomes
            # inefficient. So, we emulate its logic.

            with fits.open(fits_path) as hdu_list:
                for idx, hdu in enumerate(hdu_list):
                    if idx == 0:
                        header0 = hdu.header
                    else:
                        hdr = hdu.header
                        hdr.extend(header0, unique=True)

                        # This ccddata function often generates annoying warnings
                        with warnings.catch_warnings():
                            warnings.simplefilter('ignore')
                            hdr, wcs = ccddata._generate_wcs_and_update_header(hdr)

                        # Note: we skip all the unit-handling logic here since the LSST
                        # sim data I'm using don't have anything useful.

                        if actually_load_data:
                            data = hdu.data
                        else:
                            data = np.empty(hdu.shape, dtype=np.void)

                        ccd = ccddata.CCDData(data, meta=hdr, unit=unit, wcs=wcs)

                        ccd = ccdproc.trim_image(ccd, fits_section=ccd.header['DATASEC'])
                        yield (f'{fits_path}:{idx}', ccd)

    return loader_generator


class MultiWcsDescriptor(object):
    ident = None
    in_shape = None
    in_wcs = None

    imin = None
    imax = None
    jmin = None
    jmax = None

    sub_tiling = None


class MultiWcsProcessor(object):
    def __init__(self, loader_generator):
        self._loader_generator = loader_generator


    def compute_global_pixelization(self):
        from reproject.mosaicking.wcs_helpers import find_optimal_celestial_wcs

        # Load up current WCS information for all of the inputs

        def create_descriptor(loader_data):
            desc = MultiWcsDescriptor()
            desc.ident = loader_data[0]
            desc.in_shape = loader_data[1].shape
            desc.in_wcs = loader_data[1].wcs
            return desc

        self._descs = [create_descriptor(tup) for tup in self._loader_generator(False)]

        # Compute the optimal tangential tiling that fits all of them.

        self._combined_wcs, self._combined_shape = find_optimal_celestial_wcs(
            ((desc.in_shape, desc.in_wcs) for desc in self._descs),
            auto_rotate = True,
            projection = 'TAN',
        )

        self._tiling = StudyTiling(self._combined_shape[1], self._combined_shape[0])

        # While we're here, figure out how each input will map onto the global
        # tiling. This makes sure that nothing funky happened during the
        # computation and allows us to know how many tiles we'll have to visit.

        self._n_todo = 0

        for desc in self._descs:
            # XXX: this functionality is largely copied from
            # `reproject.mosaicking.coadd.reproject_and_coadd`, and redundant
            # with it, but it's sufficiently different that I think the best
            # approach is to essentially fork the implementation.

            # Figure out where this array lands in the mosaic.

            ny, nx = desc.in_shape
            xc = np.array([-0.5, nx - 0.5, nx - 0.5, -0.5])
            yc = np.array([-0.5, -0.5, ny - 0.5, ny - 0.5])
            xc_out, yc_out = self._combined_wcs.world_to_pixel(desc.in_wcs.pixel_to_world(xc, yc))

            if np.any(np.isnan(xc_out)) or np.any(np.isnan(yc_out)):
                raise Exception(f'segment {desc.ident} does not fit within the global mosaic')

            desc.imin = max(0, int(np.floor(xc_out.min() + 0.5)))
            desc.imax = min(self._combined_shape[1], int(np.ceil(xc_out.max() + 0.5)))
            desc.jmin = max(0, int(np.floor(yc_out.min() + 0.5)))
            desc.jmax = min(self._combined_shape[0], int(np.ceil(yc_out.max() + 0.5)))

            # Compute the sub-tiling now so that we can count how many total
            # tiles we'll need to process. Note that the combined WCS coordinate
            # system has y=0 on the bottom, whereas the tiling coordinate system
            # has y=0 at the top. So we need to invert the coordinates
            # vertically when determining the sub-tiling.

            if desc.imax < desc.imin or desc.jmax < desc.jmin:
                raise Exception(f'segment {desc.ident} maps to zero size in the global mosaic')

            desc.sub_tiling = self._tiling.compute_for_subimage(
                desc.imin,
                self._combined_shape[0] - desc.jmax,
                desc.imax - desc.imin,
                desc.jmax - desc.jmin,
            )

            self._n_todo += desc.sub_tiling.count_populated_positions()

        return self  # chaining convenience


    def tile(self, pio, reproject_function, parallel=None, cli_progress=False, **kwargs):
        """
        Tile!!!!

        Parameters
        ----------
        pio : :class:`toasty.pyramid.PyramidIO`
            A :class:`~toasty.pyramid.PyramidIO` instance to manage the I/O with
            the tiles in the tile pyramid.
        reproject_function : TKTK
            TKTK
        parallel : integer or None (the default)
            The level of parallelization to use. If unspecified, defaults to using
            all CPUs. If the OS does not support fork-based multiprocessing,
            parallel processing is not possible and serial processing will be
            forced. Pass ``1`` to force serial processing.
        cli_progress : optional boolean, defaults False
            If true, a progress bar will be printed to the terminal using tqdm.

        """
        from .par_util import resolve_parallelism
        parallel = resolve_parallelism(parallel)

        if parallel > 1:
            self._tile_parallel(pio, reproject_function, cli_progress, parallel, **kwargs)
        else:
            self._tile_serial(pio, reproject_function, cli_progress, **kwargs)


    def _tile_serial(self, pio, reproject_function, cli_progress, **kwargs):
        with tqdm(total=self._n_todo, disable=not cli_progress) as progress:
            for (ident, ccd), desc in zip(self._loader_generator(True), self._descs):
                # XXX: more copying from
                # `reproject.mosaicking.coadd.reproject_and_coadd`.

                wcs_out_indiv = self._combined_wcs[desc.jmin:desc.jmax, desc.imin:desc.imax]
                shape_out_indiv = (desc.jmax - desc.jmin, desc.imax - desc.imin)

                array = reproject_function(
                    (ccd.data, ccd.wcs),
                    output_projection=wcs_out_indiv,
                    shape_out=shape_out_indiv,
                    return_footprint=False,
                    **kwargs
                )

                # Once again, FITS coordinates have y=0 at the bottom and our
                # coordinates have y=0 at the top, so we need a vertical flip.
                image = Image.from_array(array.astype(np.float32)[::-1])

                for pos, width, height, image_x, image_y, tile_x, tile_y in desc.sub_tiling.generate_populated_positions():
                    iy_idx = slice(image_y, image_y + height)
                    ix_idx = slice(image_x, image_x + width)
                    by_idx = slice(tile_y, tile_y + height)
                    bx_idx = slice(tile_x, tile_x + width)

                    with pio.update_image(pos, masked_mode=image.mode, default='masked') as basis:
                        image.update_into_maskable_buffer(basis, iy_idx, ix_idx, by_idx, bx_idx)

                    progress.update(1)

        if cli_progress:
            print()


    def _tile_parallel(self, pio, reproject_function, cli_progress, parallel, **kwargs):
        import multiprocessing as mp

        # Start up the workers

        queue = mp.Queue(maxsize = 2 * parallel)
        workers = []

        for _ in range(parallel):
            w = mp.Process(target=_mp_tile_worker, args=(queue, pio, reproject_function, kwargs))
            w.daemon = True
            w.start()
            workers.append(w)

        # Send out them segments

        with tqdm(total=len(self._descs), disable=not cli_progress) as progress:
            for (ident, ccd), desc in zip(self._loader_generator(True), self._descs):
                wcs_out_indiv = self._combined_wcs[desc.jmin:desc.jmax, desc.imin:desc.imax]
                queue.put((ident, ccd, desc, wcs_out_indiv))
                progress.update(1)

            queue.close()

            for w in workers:
                w.join()

        if cli_progress:
            print()


def _mp_tile_worker(queue, pio, reproject_function, kwargs):
    """
    Generate and enqueue the tiles that need to be processed.
    """
    from queue import Empty

    while True:
        try:
            # un-pickling WCS objects always triggers warnings right now
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                ident, ccd, desc, wcs_out_indiv = queue.get(True, timeout=1)
        except (OSError, ValueError, Empty):
            # OSError or ValueError => queue closed. This signal seems not to
            # cross multiprocess lines, though.
            break

        shape_out_indiv = (desc.jmax - desc.jmin, desc.imax - desc.imin)

        array = reproject_function(
            (ccd.data, ccd.wcs),
            output_projection=wcs_out_indiv,
            shape_out=shape_out_indiv,
            return_footprint=False,
            **kwargs
        )

        # Once again, FITS coordinates have y=0 at the bottom and our
        # coordinates have y=0 at the top, so we need a vertical flip.
        image = Image.from_array(array.astype(np.float32)[::-1])

        for pos, width, height, image_x, image_y, tile_x, tile_y in desc.sub_tiling.generate_populated_positions():
            iy_idx = slice(image_y, image_y + height)
            ix_idx = slice(image_x, image_x + width)
            by_idx = slice(tile_y, tile_y + height)
            bx_idx = slice(tile_x, tile_x + width)

            with pio.update_image(pos, masked_mode=image.mode, default='masked') as basis:
                image.update_into_maskable_buffer(basis, iy_idx, ix_idx, by_idx, bx_idx)



# -*- mode: python; coding: utf-8 -*-
# Copyright 2013-2020 Chris Beaumont and the AAS WorldWide Telescope project
# Licensed under the MIT License.

__all__ = '''
assert_xml_elements_equal
check_xml_elements_equal
test_path
'''.split()

import os.path


TESTS_DIR = os.path.dirname(os.path.abspath(__file__))

def test_path(*pieces):
    return os.path.join(TESTS_DIR, *pieces)


def check_xml_elements_equal(observed, expected):
    """See if two XML elements are equal through recursive comparison. We do *not*
    check the "tail" text item, and we strip whitespace in "text".

    Derived om `StackExchange <https://stackoverflow.com/a/24349916/3760486>`_.

    """
    if observed.tag != expected.tag:
        return 'expected tag {}, observed tag {}'.format(expected.tag, observed.tag)

    if observed.text is None:
        otext = ''
    else:
        otext = observed.text.strip()

    if expected.text is None:
        etext = ''
    else:
        etext = expected.text.strip()

    if otext != etext:
        return 'expected text {!r} in tag {}, observed {!r}'.format(etext, expected.tag, otext)
    if observed.attrib != expected.attrib:
        return 'expected attrs {!r} in tag {}, observed {!r}'.format(expected.attrib, expected.tag, observed.attrib)
    if len(observed) != len(expected):
        return 'expected {} children of in tag {}, observed {}'.format(len(expected), expected.tag, len(observed))

    for c1, c2 in zip(observed, expected):
        reason = check_xml_elements_equal(c1, c2)
        if reason is not None:
            return reason

    return None

def assert_xml_elements_equal(observed, expected):
    reason = check_xml_elements_equal(observed, expected)
    if reason is not None:
        raise Exception('unequal XML elements: {}'.format(reason))



# -*- mode: python; coding: utf-8 -*-
# Copyright 2020 the AAS WorldWide Telescope project
# Licensed under the MIT License.

"""
Low-level loading and handling of images.

Here, images are defined as 2D data buffers stored in memory. Images might be
small, if dealing with an individual tile, or extremely large, if loading up a
large study for tiling.

"""
from __future__ import absolute_import, division, print_function

__all__ = '''
Image
ImageLoader
ImageMode
'''.split()

from enum import Enum
from PIL import Image as pil_image
import numpy as np
import sys

try:
    from astropy.io import fits
    ASTROPY_INSTALLED = True
except ImportError:
    ASTROPY_INSTALLED = False

PIL_RGB_FORMATS = {'jpg': 'JPEG'}
PIL_RGBA_FORMATS = {'png': 'PNG'}
PIL_FORMATS = PIL_RGB_FORMATS.copy()
PIL_FORMATS.update(PIL_RGBA_FORMATS)
SUPPORTED_FORMATS = list(PIL_RGB_FORMATS) + list(PIL_RGBA_FORMATS) + ['npy']

if ASTROPY_INSTALLED:
    SUPPORTED_FORMATS += ['fits']


def _validate_format(name, fmt):
    if fmt is not None and fmt not in SUPPORTED_FORMATS:
        raise ValueError('{0} should be one of {1}'.format(name, '/'.join(sorted(SUPPORTED_FORMATS))))


def _array_to_mode(array):
    if array.ndim == 2 and array.dtype.kind == 'f':
        if array.dtype.itemsize == 4:
            return ImageMode.F32
        elif array.dtype.itemsize == 8:
            return ImageMode.F64
    elif array.ndim == 3:
        if array.shape[2] == 3:
            if array.dtype.kind == 'f' and array.itemsize == 2:
                return ImageMode.F16x3
            elif array.dtype.kind == 'u' and array.dtype.itemsize == 1:
                return ImageMode.RGB
        elif array.shape[2] == 4:
            if array.dtype.kind == 'u' and array.dtype.itemsize == 1:
                return ImageMode.RGBA
    raise ValueError('Could not determine mode for array with dtype {0} and shape {1}'.format(array.dtype, array.shape))


class ImageMode(Enum):
    """
    Allowed image "modes", describing their pixel data formats.

    These align with PIL modes when possible, but we expect to need to support
    modes that aren't present in PIL (namely, float64). There are also various
    obscure PIL modes that we do not support.

    """
    RGB = 'RGB'
    RGBA = 'RGBA'
    F32 = 'F'
    F64 = 'D'
    F16x3 = 'F16x3'

    def make_maskable_buffer(self, buf_height, buf_width):
        """
        Return a new, uninitialized buffer of the specified shape, with a mode
        compatible with this one but able to accept undefined values.

        Parameters
        ----------
        buf_height : int
            The height of the new buffer
        buf_width : int
            The width of the new buffer

        Returns
        -------
        An uninitialized :class:`Image` instance.

        Notes
        -----
        "Maskable" means that the buffer can accommodate undefined values.
        If the image is RGB or RGBA, that means that the buffer will have an
        alpha channel. If the image is scientific, that means that the buffer
        will be able to accept NaNs.

        """

        if self in (ImageMode.RGB, ImageMode.RGBA):
            arr = np.empty((buf_height, buf_width, 4), dtype=np.uint8)
        elif self == ImageMode.F32:
            arr = np.empty((buf_height, buf_width), dtype=np.float32)
        elif self == ImageMode.F64:
            arr = np.empty((buf_height, buf_width), dtype=np.float64)
        elif self == ImageMode.F16x3:
            arr = np.empty((buf_height, buf_width, 3), dtype=np.float16)
        else:
            raise Exception('unhandled mode in make_maskable_buffer()')

        return Image.from_array(arr)

    def try_as_pil(self):
        """
        Attempt to convert this mode into a PIL image mode string.

        Returns
        -------
        A PIL image mode string, or None if there is no exact counterpart.
        """
        if self == ImageMode.F16x3:
            return None
        return self.value


class ImageLoader(object):
    """
    A class defining how to load an image.

    This is implemented as its own class since there can be some options
    involved, and we want to provide a centralized place for handling them all.
    """
    black_to_transparent = False
    colorspace_processing = 'srgb'
    crop = None
    psd_single_layer = None

    @classmethod
    def add_arguments(cls, parser):
        """
        Add standard image-loading options to an argparse parser object.

        Parameters
        ----------
        parser : :class:`argparse.ArgumentParser`
            The argument parser to modify

        Returns
        -------
        The :class:`ImageLoader` class (for chainability).

        Notes
        -----
        If you are writing a command-line interface that takes a single image as
        an input, use this function to wire in to standardized image-loading
        infrastructure and options.

        """
        parser.add_argument(
            '--black-to-transparent',
            action = 'store_true',
            help = 'Convert full black colors to be transparent',
        )
        parser.add_argument(
            '--colorspace-processing',
            metavar = 'MODE',
            default = 'srgb',
            help = 'What kind of RGB colorspace processing to perform (default: %(default)s; choices: %(choices)s)',
            choices = ['srgb', 'none'],
        )
        parser.add_argument(
            '--crop',
            metavar = 'TOP,RIGHT,BOTTOM,LEFT',
            help = 'Crop the input image by discarding pixels from each edge (default: 0,0,0,0)',
        )
        parser.add_argument(
            '--psd-single-layer',
            type = int,
            metavar = 'NUMBER',
            help = 'If loading a Photoshop image, the (0-based) layer number to load -- saves memory',
        )
        return cls

    @classmethod
    def create_from_args(cls, settings):
        """
        Process standard image-loading options to create an :class:`ImageLoader`.

        Parameters
        ----------
        settings : :class:`argparse.Namespace`
            Settings from processing command-line arguments

        Returns
        -------
        A new :class:`ImageLoader` initialized with the settings.
        """
        loader = cls()
        loader.black_to_transparent = settings.black_to_transparent
        loader.colorspace_processing = settings.colorspace_processing
        loader.psd_single_layer = settings.psd_single_layer

        if settings.crop is not None:
            try:
                crop = list(map(int, settings.crop.split(',')))
                assert all(c >= 0 for c in crop)
                assert len(crop) in (1, 2, 4)
            except Exception:
                raise Exception('cannot parse `--crop` setting `{settings.crop!r}`: should be a comma-separated list of 1, 2, or 4 non-negative integers')

            if len(crop) == 1:
                c = crop[0]
                crop = [c, c, c, c]
            elif len(crop) == 2:
                cv, ch = crop
                crop = [cv, ch, cv, ch]

            loader.crop = crop

        return loader

    def load_pil(self, pil_img):
        """
        Load an already opened PIL image.

        Parameters
        ----------
        pil_img : :class:`PIL.Image.Image`
            The image.

        Returns
        -------
        A new :class:`Image`.

        Notes
        -----
        This function should be used instead of :meth:`Image.from_pil` because
        may postprocess the image in various ways, depending on the loader
        configuration.

        """
        # If we're cropping, do it.
        if self.crop is not None:
            upper = self.crop[0]
            right = pil_img.width - self.crop[1]
            lower = pil_img.height - self.crop[2]
            left = self.crop[3]

            # This operation can trigger PIL decompression-bomb aborts, so
            # avoid them in the standard, non-thread-safe, way.
            old_max = pil_image.MAX_IMAGE_PIXELS
            try:
                pil_image.MAX_IMAGE_PIXELS = None
                pil_img = pil_img.crop((left, upper, right, lower))
            finally:
                pil_image.MAX_IMAGE_PIXELS = old_max

        # If an unrecognized mode, try to standardize it. The weird PIL modes
        # don't generally support an alpha channel, so we convert to RGB.

        try:
            ImageMode(pil_img.mode)
        except ValueError:
            print('warning: trying to convert image file to RGB from unexpected bitmode "%s"' % pil_img.mode)

            if self.black_to_transparent:
                # Avoid double-converting in the next filter.
                pil_img = pil_img.convert('RGBA')
            else:
                pil_img = pil_img.convert('RGB')

        # Convert pure black to transparent -- make sure to do this before any
        # colorspace processing.
        #
        # As far as I can tell, PIL has no good way to modify an image on a
        # pixel-by-pixel level in-place, which is really annoying. For now I'm
        # doing this processing in Numpy space, but for an image that almost
        # fills up memory we might have to use a different approach since this
        # one will involve holding two buffers at once.

        if self.black_to_transparent:
            if pil_img.mode != 'RGBA':
                pil_img = pil_img.convert('RGBA')
            a = np.asarray(pil_img)
            a = a.copy()  # read-only buffer => writeable

            for i in range(a.shape[0]):
                nonblack = (a[i,...,0] > 0)
                np.logical_or(nonblack, a[i,...,1] > 0, out=nonblack)
                np.logical_or(nonblack, a[i,...,2] > 0, out=nonblack)
                a[i,...,3] *= nonblack

            # This is my attempt to preserve the image metadata and other
            # attributes, swapping out the pixel data only. There is probably
            # a better way to do this
            new_img = pil_image.fromarray(a, mode=pil_img.mode)
            pil_img.im = new_img.im
            del a, new_img

        # Make sure that we end up in the right color space. From experience, some
        # EPO images have funky colorspaces and we need to convert to sRGB to get
        # the tiled versions to appear correctly.

        if self.colorspace_processing != 'none' and 'icc_profile' in pil_img.info:
            assert self.colorspace_processing == 'srgb' # more modes, one day?

            try:
                from PIL import ImageCms
                # ImageCms doesn't raise import error if the implementation is unavailable
                # "for doc purposes". To see if it's available we need to actually try to
                # do something:
                out_prof = ImageCms.createProfile('sRGB')
            except ImportError:
                print('''warning: colorspace processing requested, but no `ImageCms` module found in PIL.
    Your installation of PIL probably does not have colorspace support.
    Colors will not be transformed to sRGB and therefore may not appear as intended.
    Compare toasty's output to your source image and decide if this is acceptable to you.
    Consider a different setting of the `--colorspace-processing` argument to avoid this warning.''',
                    file=sys.stderr)
            else:
                from io import BytesIO
                in_prof = ImageCms.getOpenProfile(BytesIO(pil_img.info['icc_profile']))
                xform = ImageCms.buildTransform(in_prof, out_prof, pil_img.mode, pil_img.mode)
                ImageCms.applyTransform(pil_img, xform, inPlace=True)

        return Image.from_pil(pil_img)

    def load_stream(self, stream):
        """
        Load an image into memory from a file-like stream.

        Parameters
        ----------
        stream : file-like
            The data to load. Reads should yield bytes.

        Returns
        -------
        A new :class:`Image`.

        """
        # TODO: one day, we'll support FITS files and whatnot and we'll have a
        # mode where we get a Numpy array but not a PIL image. For now, just
        # pass it off to PIL and hope for the best.

        # Prevent PIL decompression-bomb aborts. Not thread-safe, of course.
        old_max = pil_image.MAX_IMAGE_PIXELS

        try:
            pil_image.MAX_IMAGE_PIXELS = None
            pilimg = pil_image.open(stream)
        finally:
            pil_image.MAX_IMAGE_PIXELS = old_max

        # Now pass it off to generic PIL handling ...
        return self.load_pil(pilimg)

    def load_path(self, path):
        """
        Load an image into memory from a filesystem path.

        Parameters
        ----------
        path : str
            The filesystem path to load.

        Returns
        -------
        A new :class:`Image`.
        """
        # Special handling for Numpy arrays. TODO: it would be better to sniff
        # filetypes instead of just looking at extensions. But, lazy.

        if path.endswith('.npy'):
            arr = np.load(path)
            return Image.from_array(arr, default_format='npy')

        if path.lower().endswith(('.fits', '.fts', '.fits.gz', '.fts.gz')):

            # TODO: implement a better way to recognize FITS files
            # TODO: decide how to handle multiple HDUs
            # TODO: decide how to handle non-TAN projections

            hdu = fits.open(path)[0]
            arr = hdu.data
            if ASTROPY_INSTALLED:
                from astropy.wcs import WCS
                wcs = WCS(hdu.header)
            else:
                wcs = None

            return Image.from_array(arr, wcs=wcs, default_format='fits')

        # Special handling for Photoshop files, used for some very large mosaics
        # with transparency (e.g. the PHAT M31/M33 images).

        # TODO: check for AVM in following formats and set WCS using this if needed.

        if path.endswith('.psd') or path.endswith('.psb'):
            try:
                from psd_tools import PSDImage
            except ImportError:
                pass
            else:
                psd = PSDImage.open(path)

                # If the Photoshop image is a single layer, we can save a lot of
                # memory by not using the composite() function. This has helped
                # me process very large Photoshop files.
                if self.psd_single_layer is not None:
                    pilimg = psd[self.psd_single_layer].topil()
                else:
                    pilimg = psd.composite()

                return self.load_pil(pilimg)

        # Special handling for OpenEXR files, used for large images with high
        # dynamic range.

        if path.endswith('.exr'):
            from .openexr import load_openexr
            img = load_openexr(path)
            if img.dtype != np.float16:
                raise Exception('only half-precision OpenEXR images are currently supported')
            return Image.from_array(img)

        # (One day, maybe we'll do more kinds of sniffing.) No special handling
        # came into play; just open the file and auto-detect.

        with open(path, 'rb') as f:
            return self.load_stream(f)


class Image(object):
    """A 2D data array stored in memory.

    This class primarily exists to help us abstract between the cases where we
    have "bitmap" RGB(A) images and "science" floating-point images.

    """
    _pil = None
    _array = None
    _mode = None
    _default_format = 'png'
    _wcs = None

    @classmethod
    def from_pil(cls, pil_img, wcs=None, default_format=None):
        """
        Create a new Image from a PIL image.

        Parameters
        ----------
        pil_img : :class:`PIL.Image.Image`
            The source image.
        wcs : :class:`~astropy.wcs.WCS`, optional
            The WCS coordinate system for the image.
        default_format : str, optional
            The default format to use when writing the image if none is
            specified explicitly.

        Returns
        -------
        A new :class:`Image` wrapping the PIL image.
        """

        _validate_format('default_format', default_format)

        # Make sure that the image data are actually loaded from disk. Pillow
        # lazy-loads such that sometimes `np.asarray(img)` ends up failing
        # mysteriously if the image is loaded from a file handle that is closed
        # promptly.
        pil_img.load()

        inst = cls()
        inst._pil = pil_img
        inst._wcs = wcs
        inst._default_format = default_format or cls._default_format

        try:
            inst._mode = ImageMode(pil_img.mode)
        except ValueError:
            raise Exception('image mode {} is not supported'.format(pil_img.mode))

        return inst

    @classmethod
    def from_array(cls, array, wcs=None, default_format=None):
        """Create a new Image from an array-like data variable.

        Parameters
        ----------
        array : array-like object
            The source data. This should either be a 2-d floating point array a
            3-d floating-point or uint8 array with shape (3, ny, nx), or a 3-d
            uint8 array with shape (4, ny, nx).
        wcs : :class:`~astropy.wcs.WCS`, optional
            The WCS coordinate system for the image.
        default_format : str, optional
            The default format to use when writing the image if none is
            specified explicitly. If not specified, this is automatically
            chosen at write time based on the array type.

        Returns
        -------
        A new :class:`Image` wrapping the data.

        Notes
        -----
        The array will be converted to be at least two-dimensional.

        """

        _validate_format('default_format', default_format)

        array = np.atleast_2d(array)

        inst = cls()
        inst._mode = _array_to_mode(array)
        inst._default_format = default_format or cls._default_format
        inst._array = array
        inst._wcs = wcs
        return inst

    def asarray(self):
        """Obtain the image data as a Numpy array.

        Returns
        -------
        If the image is an RGB(A) bitmap, the array will have shape ``(height, width, planes)``
        and a dtype of ``uint8``, where ``planes`` is either 3
        or 4 depending on whether the image has an alpha channel. If the image
        is science data, it will have shape ``(height, width)`` and have a
        floating-point dtype.

        """
        # NOTE: it turns out that np.asarray() on a PIL image has to copy the
        # entire image data. So, on a large image, it becomes super slow.
        # Therefore we cache the array. The array is marked read-only so we
        # don't have to worry about it and the PIL image getting out of sync
        # with modifications.
        if self._array is None:
            self._array = np.asarray(self._pil)
        return self._array

    def aspil(self):
        """Obtain the image data as :class:`PIL.Image.Image`.

        Returns
        -------
        If the image was loaded as a PIL image, the underlying object will be
        returned. Otherwise the data array will be converted into a PIL image,
        which requires that the array have an RGB(A) format with a shape of
        ``(height, width, planes)``, where ``planes`` is 3 or 4, and a dtype of
        ``uint8``.

        """
        if self._pil is not None:
            return self._pil
        if self.mode.try_as_pil() is None:
            raise Exception(f'Toasty image with mode {self.mode} cannot be converted to PIL')
        return pil_image.fromarray(self._array)

    @property
    def mode(self):
        return self._mode

    @property
    def dtype(self):
        # TODO: can this be more efficient? Does it need to be?
        return self.asarray().dtype

    @property
    def wcs(self):
        return self._wcs

    @property
    def shape(self):
        if self._array is not None:
            return self._array.shape

        return (self._pil.height, self._pil.width, len(self._pil.getbands()))

    @property
    def width(self):
        return self.shape[1]

    @property
    def height(self):
        return self.shape[0]

    @property
    def default_format(self):
        if self._default_format is None:
            if self.mode in (ImageMode.RGB, ImageMode.RGBA):
                return 'png'
            elif self.mode in (ImageMode.F32, ImageMode.F16x3):
                return 'npy'
        else:
            return self._default_format

    @default_format.setter
    def default_format(self, value):
        if value in SUPPORTED_FORMATS:
            self._default_format = value
        else:
            raise ValueError('Unrecognized format: {0}'.format(value))

    def fill_into_maskable_buffer(self, buffer, iy_idx, ix_idx, by_idx, bx_idx):
        """
        Fill a maskable buffer with a rectangle of data from this image.

        Parameters
        ----------
        buffer : :class:`Image`
            The destination buffer image, created with :meth:`ImageMode.make_maskable_buffer`.
        iy_idx : slice or other indexer
            The indexer into the Y axis of the source image (self).
        ix_idx : slice or other indexer
            The indexer into the X axis of the source image (self).
        by_idx : slice or other indexer
            The indexer into the Y axis of the destination *buffer*.
        bx_idx : slice or other indexer
            The indexer into the X axis of the destination *buffer*.

        Notes
        -----
        This highly specialized function is used to tile images efficiently. No
        bounds checking is performed. The rectangles defined by the indexers in
        the source and destination are assumed to agree in size. The regions of
        the buffer not filled by source data are masked, namely: either filled
        with alpha=0 or with NaN, depending on the image mode.

        """
        i = self.asarray()
        b = buffer.asarray()

        if self.mode == ImageMode.RGB:
            b.fill(0)
            b[by_idx,bx_idx,:3] = i[iy_idx,ix_idx]
            b[by_idx,bx_idx,3] = 255
        elif self.mode == ImageMode.RGBA:
            b.fill(0)
            b[by_idx,bx_idx] = i[iy_idx,ix_idx]
        elif self.mode in (ImageMode.F32, ImageMode.F16x3):
            b.fill(np.nan)
            b[by_idx,bx_idx] = i[iy_idx,ix_idx]
        else:
            raise Exception('unhandled mode in fill_into_maskable_buffer')

    def update_into_maskable_buffer(self, buffer, iy_idx, ix_idx, by_idx, bx_idx):
        """
        Update a maskable buffer with data from this image.

        Parameters
        ----------
        buffer : :class:`Image`
            The destination buffer image, created with :meth:`ImageMode.make_maskable_buffer`.
        iy_idx : slice or other indexer
            The indexer into the Y axis of the source image (self).
        ix_idx : slice or other indexer
            The indexer into the X axis of the source image (self).
        by_idx : slice or other indexer
            The indexer into the Y axis of the destination *buffer*.
        bx_idx : slice or other indexer
            The indexer into the X axis of the destination *buffer*.

        Notes
        -----
        Unlike :meth:`fill_into_maskable_buffer`, this function does not clear
        the entire buffer. It only overwrites the portion of the buffer covered
        by non-NaN-like values of the input image.

        """
        i = self.asarray()
        b = buffer.asarray()

        sub_b = b[by_idx,bx_idx]
        sub_i = i[iy_idx,ix_idx]

        if self.mode == ImageMode.RGB:
            sub_b[...,:3] = sub_i
            sub_b[...,3] = 255
        elif self.mode == ImageMode.RGBA:
            valid = (sub_i[...,3] != 0)
            valid = np.broadcast_to(valid[...,None], sub_i.shape)
            np.putmask(sub_b, valid, sub_i)
        elif self.mode == ImageMode.F32:
            valid = ~np.isnan(sub_i)
            np.putmask(sub_b, valid, sub_i)
        elif self.mode == ImageMode.F16x3:
            valid = ~np.any(np.isnan(sub_i), axis=2)
            valid = np.broadcast_to(valid[...,None], sub_i.shape)
            np.putmask(sub_b, valid, sub_i)
        else:
            raise Exception('unhandled mode in update_into_maskable_buffer')

    def save(self, path_or_stream, format=None, mode=None):
        """
        Save this image to a filesystem path or stream

        Parameters
        ----------
        path_or_stream : path-like object or file-like object
            The destination into which the data should be written. If file-like,
            the stream should accept bytes.

        """

        _validate_format('format', format)

        format = format or self._default_format

        if format in PIL_FORMATS:
            pil_image = self.aspil()
            if format in PIL_RGB_FORMATS and mode is None:
                mode = ImageMode.RGB
            if mode is not None:
                pil_image = pil_image.convert(mode.try_as_pil())
            pil_image.save(path_or_stream, format=PIL_FORMATS[format])
        elif format == 'npy':
            np.save(path_or_stream, self.asarray())
        elif format == 'fits':
            if self._wcs is None:
                header = None if self._wcs is None else self._wcs.toheader()
            fits.writeto(path_or_stream, self.asarray(),
                         header=header, overwrite=True)

    def make_thumbnail_bitmap(self):
        """Create a thumbnail bitmap from the image.

        Returns
        -------
        An RGB :class:`PIL.Image.Image` representing a thumbnail of the input
        image. WWT thumbnails are 96 pixels wide and 45 pixels tall and should
        be saved in JPEG format.

        """
        if self.mode in (ImageMode.F32, ImageMode.F64, ImageMode.F16x3):
            raise Exception('cannot thumbnail-ify non-RGB Image')

        THUMB_SHAPE = (96, 45)
        THUMB_ASPECT = THUMB_SHAPE[0] / THUMB_SHAPE[1]

        if self.width / self.height > THUMB_ASPECT:
            # The image is wider than desired; we'll need to crop off the sides.
            target_width = int(round(self.height * THUMB_ASPECT))
            dx = (self.width - target_width) // 2
            crop_box = (dx, 0, dx + target_width, self.height)
        else:
            # The image is taller than desired; crop off top and bottom.
            target_height = int(round(self.width / THUMB_ASPECT))
            dy = (self.height - target_height) // 2
            crop_box = (0, dy, self.width, dy + target_height)

        # Turns out that PIL the decompression-bomb checks happen here too.
        # Disable in our usual, not thread-safe, way.

        old_max = pil_image.MAX_IMAGE_PIXELS

        try:
            pil_image.MAX_IMAGE_PIXELS = None
            thumb = self.aspil().crop(crop_box)
        finally:
            pil_image.MAX_IMAGE_PIXELS = old_max

        thumb.thumbnail(THUMB_SHAPE)

        # Depending on the source image, the mode might be RGBA, which can't
        # be JPEG-ified.
        thumb = thumb.convert('RGB')

        return thumb

    def clear(self):
        """
        Fill the image with whatever "empty" value is most appropriate for its
        mode.

        Notes
        -----
        The image is assumed to be writable, which will not be the case for
        images constructed from PIL. If the mode is RGB or RGBA, the buffer is
        filled with zeros. If the mode is floating-point, the buffer is filled
        with NaNs.

        """
        if self._mode in (ImageMode.RGB, ImageMode.RGBA):
            self.asarray().fill(0)
        elif self._mode in (ImageMode.F32, ImageMode.F64, ImageMode.F16x3):
            self.asarray().fill(np.nan)
        else:
            raise Exception('unhandled mode in clear()')



# -*- mode: python; coding: utf-8 -*-
# Copyright 2019-2020 the AAS WorldWide Telescope project.
# Licensed under the MIT License.

"""
Entrypoints for the "toasty pipeline" command-line tools.
"""

__all__ = '''
pipeline_getparser
pipeline_impl
'''.split()

import argparse
from fnmatch import fnmatch
import glob
import os.path
import sys
from wwt_data_formats.cli import EnsureGlobsExpandedAction

from ..cli import die, warn
from . import NotActionableError


def evaluate_imageid_args(searchdir, args):
    """
    Figure out which image-ID's to process.
    """

    matched_ids = set()
    globs_todo = set()

    for arg in args:
        if glob.has_magic(arg):
            globs_todo.add(arg)
        else:
            # If an ID is explicitly (non-gobbily) added, always add it to the
            # list, without checking if it exists in `searchdir`. We could check
            # for it in searchdir now, but we'll have to check later anyway, so
            # we don't bother.
            matched_ids.add(arg)

    if len(globs_todo):
        for filename in os.listdir(searchdir):
            for g in globs_todo:
                if fnmatch(filename, g):
                    matched_ids.add(filename)
                    break

    return sorted(matched_ids)


# The "approve" subcommand

def approve_setup_parser(parser):
    parser.add_argument(
        '--workdir',
        metavar = 'PATH',
        default = '.',
        help = 'The working directory for this processing session'
    )
    parser.add_argument(
        'cand_ids',
        nargs = '+',
        action = EnsureGlobsExpandedAction,
        metavar = 'IMAGE-ID',
        help = 'Name(s) of image(s) to approve for publication (globs accepted)'
    )


def approve_impl(settings):
    from wwt_data_formats.folder import Folder, make_absolutizing_url_mutator
    from . import PipelineManager

    mgr = PipelineManager(settings.workdir)
    mgr.ensure_config()

    pub_url_prefix = mgr._config.get('publish_url_prefix')
    if pub_url_prefix:
        if pub_url_prefix[-1] != '/':
            pub_url_prefix += '/'

    proc_dir = mgr._ensure_dir('processed')
    app_dir = mgr._ensure_dir('approved')

    for cid in evaluate_imageid_args(proc_dir, settings.cand_ids):
        if not os.path.isdir(os.path.join(proc_dir, cid)):
            die(f'no such processed candidate ID {cid!r}')

        index_path = os.path.join(proc_dir, cid, 'index.wtml')
        prefix = pub_url_prefix + cid + '/'

        try:
            f = Folder.from_file(os.path.join(proc_dir, cid, 'index_rel.wtml'))
            f.mutate_urls(make_absolutizing_url_mutator(prefix))

            with open(index_path, 'wt', encoding='utf8') as f_out:
                f.write_xml(f_out)
        except Exception as e:
            print('error: failed to create index.wtml from index_rel.wtml', file=sys.stderr)

            try:
                os.remove(index_path)
            except Exception:
                pass

            raise

        os.rename(os.path.join(proc_dir, cid), os.path.join(app_dir, cid))


# The "fetch" subcommand

def fetch_setup_parser(parser):
    parser.add_argument(
        '--workdir',
        metavar = 'PATH',
        default = '.',
        help = 'The working directory for this processing session'
    )
    parser.add_argument(
        'cand_ids',
        nargs = '+',
        action = EnsureGlobsExpandedAction,
        metavar = 'CAND-ID',
        help = 'Name(s) of candidate(s) to fetch and prepare for processing (globs accepted)'
    )


def fetch_impl(settings):
    from . import PipelineManager

    mgr = PipelineManager(settings.workdir)
    cand_dir = mgr._ensure_dir('candidates')
    rej_dir = mgr._ensure_dir('rejects')
    src = mgr.get_image_source()

    for cid in evaluate_imageid_args(cand_dir, settings.cand_ids):
        # Funky structure here is to try to ensure that cdata is closed in case
        # a NotActionable happens, so that we can move the directory on Windows.
        try:
            try:
                cdata = open(os.path.join(cand_dir, cid), 'rb')
            except FileNotFoundError:
                die(f'no such candidate ID {cid!r}')

            try:
                print(f'fetching {cid} ... ', end='')
                sys.stdout.flush()
                cachedir = mgr._ensure_dir('cache_todo', cid)
                src.fetch_candidate(cid, cdata, cachedir)
                print('done')
            finally:
                cdata.close()
        except NotActionableError as e:
            print('not usable:', e)
            os.rename(os.path.join(cand_dir, cid), os.path.join(rej_dir, cid))
            os.rmdir(cachedir)


# The "init" subcommand

def init_setup_parser(parser):
    parser.add_argument(
        '--azure-conn-env',
        metavar = 'ENV-VAR-NAME',
        help = 'The name of an environment variable contain an Azure Storage '
                'connection string'
    )
    parser.add_argument(
        '--azure-container',
        metavar = 'CONTAINER-NAME',
        help = 'The name of a blob container in the Azure storage account'
    )
    parser.add_argument(
        '--azure-path-prefix',
        metavar = 'PATH-PREFIX',
        help = 'A slash-separated path prefix for blob I/O within the container'
    )
    parser.add_argument(
        '--local',
        metavar = 'PATH',
        help = 'Use the local-disk I/O backend'
    )
    parser.add_argument(
        'workdir',
        nargs = '?',
        metavar = 'PATH',
        default = '.',
        help = 'The working directory for this processing session'
    )


def _pipeline_io_from_settings(settings):
    from . import azure_io, local_io

    if settings.local:
        return local_io.LocalPipelineIo(settings.local)

    if settings.azure_conn_env:
        conn_str = os.environ.get(settings.azure_conn_env)
        if not conn_str:
            die('--azure-conn-env=%s provided, but that environment variable is unset'
                % settings.azure_conn_env)

        if not settings.azure_container:
            die('--azure-container-name must be provided if --azure-conn-env is')

        path_prefix = settings.azure_path_prefix
        if not path_prefix:
            path_prefix = ''

        azure_io.assert_enabled()

        return azure_io.AzureBlobPipelineIo(
            conn_str,
            settings.azure_container,
            path_prefix
        )

    die('An I/O backend must be specified with the arguments --local or --azure-*')


def init_impl(settings):
    pipeio = _pipeline_io_from_settings(settings)
    os.makedirs(settings.workdir, exist_ok=True)
    pipeio.save_config(os.path.join(settings.workdir, 'toasty-store-config.yaml'))


# The "refresh" subcommand
#
# TODO: for large feeds, we should potentially add features to make it so that
# we don't re-check every single candidate that's ever been posted.

def refresh_setup_parser(parser):
    parser.add_argument(
        '--workdir',
        nargs = '?',
        metavar = 'PATH',
        default = '.',
        help = 'The working directory for this processing session'
    )


def refresh_impl(settings):
    from . import PipelineManager

    mgr = PipelineManager(settings.workdir)
    cand_dir = mgr._ensure_dir('candidates')
    rej_dir = mgr._ensure_dir('rejects')
    src = mgr.get_image_source()
    n_cand = 0
    n_saved = 0
    n_done = 0
    n_skipped = 0
    n_rejected = 0

    for cand in src.query_candidates():
        n_cand += 1
        uniq_id = cand.get_unique_id()

        if mgr._pipeio.check_exists(uniq_id, 'index.wtml'):
            n_done += 1
            continue  # skip already-done inputs

        if mgr._pipeio.check_exists(uniq_id, 'skip.flag'):
            n_skipped += 1
            continue  # skip inputs that are explicitly flagged

        cand_path = os.path.join(cand_dir, uniq_id)

        try:
            with open(cand_path, 'wb') as f:
                cand.save(f)
            n_saved += 1
        except NotActionableError as e:
            os.remove(cand_path)

            with open(os.path.join(rej_dir, uniq_id, 'wb')) as f:
                pass  # for now, just touch the file

            n_rejected += 1

    print(f'analyzed {n_cand} candidates from the image source')
    print(f'  - {n_saved} processing candidates saved')
    print(f'  - {n_rejected} rejected as definitely unusable')
    print(f'  - {n_done} were already done')
    print(f'  - {n_skipped} were already marked to be ignored')
    print()
    print('See the `candidates` directory for candidate image IDs.')


# Other subcommands not yet split out.

def pipeline_getparser(parser):
    subparsers = parser.add_subparsers(dest='pipeline_command')
    approve_setup_parser(subparsers.add_parser('approve'))
    fetch_setup_parser(subparsers.add_parser('fetch'))
    init_setup_parser(subparsers.add_parser('init'))

    parser = subparsers.add_parser('process-todos')
    parser.add_argument(
        '--workdir',
        nargs = '?',
        metavar = 'WORKDIR',
        default = '.',
        help = 'The local working directory',
    )

    parser = subparsers.add_parser('publish')
    parser.add_argument(
        '--workdir',
        nargs = '?',
        metavar = 'WORKDIR',
        default = '.',
        help = 'The local working directory',
    )

    refresh_setup_parser(subparsers.add_parser('refresh'))


def pipeline_impl(settings):
    from . import PipelineManager

    if settings.pipeline_command is None:
        print('Run the "pipeline" command with `--help` for help on its subcommands')
        return

    if settings.pipeline_command == 'approve':
        approve_impl(settings)
    elif settings.pipeline_command == 'fetch':
        fetch_impl(settings)
    elif settings.pipeline_command == 'init':
        init_impl(settings)
    elif settings.pipeline_command == 'process-todos':
        mgr = PipelineManager(settings.workdir)
        mgr.process_todos()
    elif settings.pipeline_command == 'publish':
        mgr = PipelineManager(settings.workdir)
        mgr.publish()
    elif settings.pipeline_command == 'refresh':
        refresh_impl(settings)
    else:
        die('unrecognized "pipeline" subcommand ' + settings.pipeline_command)



# -*- mode: python; coding: utf-8 -*-
# Copyright 2020 the AAS WorldWide Telescope project
# Licensed under the MIT License.

"""A framework for automating the ingest of source images into the formats
used by AAS WorldWide Telescope.

"""
from __future__ import absolute_import, division, print_function

__all__ = '''
CandidateInput
ImageSource
IMAGE_SOURCE_CLASS_LOADERS
NotActionableError
PipelineIo
PIPELINE_IO_LOADERS
'''.split()

from abc import ABC, abstractclassmethod, abstractmethod
from datetime import datetime, timezone
import numpy as np
import os.path
import shutil
import sys
from urllib.parse import urlsplit, quote as urlquote
from wwt_data_formats import write_xml_doc
from wwt_data_formats.folder import Folder
from wwt_data_formats.imageset import ImageSet
from wwt_data_formats.place import Place
import yaml


class NotActionableError(Exception):
    """Raised when an image is provided to the pipeline but for some reason we're
    not going to be able to get it into a WWT-compatible form.

    """
    def __init__(self, reason):
        super(NotActionableError, self).__init__(reason)


PIPELINE_IO_LOADERS = {}

def _load_local_pio(config):
    from .local_io import LocalPipelineIo
    return LocalPipelineIo._new_from_config(config)
PIPELINE_IO_LOADERS['local'] = _load_local_pio

def _load_azure_blob_pio(config):
    from .azure_io import AzureBlobPipelineIo
    return AzureBlobPipelineIo._new_from_config(config)
PIPELINE_IO_LOADERS['azure-blob'] = _load_azure_blob_pio


class PipelineIo(ABC):
    """
    An abstract base class for I/O relating to pipeline processing. An instance
    of this class might be used to fetch files from, and send them to, a cloud
    storage system like S3 or Azure Storage.
    """

    @abstractmethod
    def _export_config(self):
        """
        Export this object's configuration for serialization.

        Returns
        -------
        A dictionary of settings that can be saved as YAML format. There should
        be a key named "_type" with a string value identifying the I/O
        implementation type.
        """

    def save_config(self, path):
        """
        Save this object's configuration to the specified filesystem path.
        """
        cfg = self._export_config()

        # The config contains secrets, so create it privately and securely.
        opener = lambda path, _mode: os.open(path, os.O_WRONLY | os.O_CREAT | os.O_EXCL, mode=0o600)

        with open(path, 'wt', opener=opener, encoding='utf8') as f:
            yaml.dump(cfg, f, yaml.SafeDumper)

    @abstractclassmethod
    def _new_from_config(cls, config):
        """
        Create a new instance of this class based on serialized configuration.

        Parameters
        ----------
        config : dict
            A dict of configuration that was created with ``_export_config``

        Returns
        -------
        A new instance of the class.
        """

    @classmethod
    def load_from_config(self, path):
        """
        Create a new I/O backend from saved configuration.

        Parameters
        ----------
        path : path-like
            The path where the configuration was saved.

        Returns
        -------
        A new instance implementing the PipelineIO abstract base class.
        """

        with open(path, 'rt', encoding='utf8') as f:
            config = yaml.safe_load(f)

        ty = config.get('_type')
        loader = PIPELINE_IO_LOADERS.get(ty)
        if loader is None:
            raise Exception(f'unrecognized pipeline I/O storage type {ty!r}')

        return loader(config)

    @abstractmethod
    def check_exists(self, *path):
        """Test whether an item at the specified path exists.

        Parameters
        ----------
        *path : strings
            The path to the item, intepreted as components in a folder hierarchy.

        Returns
        -------
        A boolean indicating whether the item in question exists.

        """

    @abstractmethod
    def get_item(self, *path, dest=None):
        """Fetch a file-like item at the specified path, writing its contents into the
        specified file-like object *dest*.

        Parameters
        ----------
        *path : strings
            The path to the item, intepreted as components in a folder hierarchy.
        dest : writeable file-like object
            The object into which the item's data will be written as bytes.

        Returns
        -------
        None.

        """

    @abstractmethod
    def put_item(self, *path, source=None):
        """Put a file-like item at the specified path, reading its contents from the
        specified file-like object *source*.

        Parameters
        ----------
        *path : strings
            The path to the item, intepreted as components in a folder hierarchy.
        source : readable file-like object
            The object from which the item's data will be read, as bytes.

        Returns
        -------
        None.

        """

    @abstractmethod
    def list_items(self, *path):
        """List the items contained in the folder at the specified path.

        Parameters
        ----------
        *path : strings
            The path to the item, intepreted as components in a folder hierarchy.

        Returns
        -------
        An iterable of ``(stem, is_folder)``, where *stem* is the "basename" of an
        item contained within the specified folder and *is_folder* is a boolean
        indicating whether this item appears to be a folder itself.

        """


class ImageSource(ABC):
    """An abstract base class representing a source of images to be processed in
    the image-processing pipeline. An instance of this class might fetch
    images from an RSS feed or an AstroPix search.

    """
    @abstractclassmethod
    def get_config_key(cls):
        """Get the name of the section key used for this source's configuration data.

        Returns
        -------
        A string giving a key name usable in a YAML file.

        """

    @abstractclassmethod
    def deserialize(cls, data):
        """Create an instance of this class by deserializing configuration data.

        Parameters
        ----------
        data : dict-like object
            A dict-like object containing configuration items deserialized from
            a format such as JSON or YAML. The particular contents can vary
            depending on the implementation.

        Returns
        -------
        An instance of *cls*

        """

    @abstractmethod
    def query_candidates(self):
        """
        Generate a sequence of candidate input images that the pipeline may want
        to process.

        Returns
        -------
        A generator that yields a sequence of :class:`toasty.pipeline.CandidateInput` instances.

        """

    @abstractmethod
    def fetch_candidate(self, unique_id, cand_data_stream, cachedir):
        """
        Download a candidate image and prepare it for processing.

        Parameters
        ----------
        unique_id : str
            The unique ID returned by the :class:`toasty.pipeline.CandidateInput` instance
            that was returned from the initial query.
        cand_data_stream : readable stream returning bytes
            A data stream returning the data that were saved when the candidate
            was queried (:meth:`toasty.pipeline.CandidateInput.save`).
        cachedir : path-like
            A path pointing to a local directory inside of which the
            full image data and metadata should be cached.
        """

    @abstractmethod
    def process(self, unique_id, cand_data_stream, cachedir, builder):
        """
        Process an input into WWT format.

        Parameters
        ----------
        unique_id : str
            The unique ID returned by the :class:`toasty.pipeline.CandidateInput` instance
            that was returned from the initial query.
        cand_data_stream : readable stream returning bytes
            A data stream returning the data that were saved when the candidate
            was queried (:meth:`toasty.pipeline.CandidateInput.save`).
        cachedir : path-like
            A path pointing to a local directory inside of which the
            full image data and metadata should be cached.
        builder : :class:`toasty.builder.Builder`
            State object for constructing the WWT data files.

        Notes
        -----
        Your image processor should run the tile cascade, if needed, but the
        caller will take care of emitting the ``index_rel.wtml`` file.
        """


class CandidateInput(ABC):
    """An abstract base class representing an image from one of our sources. If it
    has not been processed before, we will fetch its data and queue it for
    processing.

    """
    @abstractmethod
    def get_unique_id(self):
        """
        Get an ID for this image that will be unique in its :class:`toasty.pipeline.ImageSource`.

        Returns
        -------
        An identifier as a string. Should be limited to path-friendly
        characters, i.e. ASCII without spaces.

        """

    @abstractmethod
    def save(self, stream):
        """
        Serialize candidate information for future processing

        Parameters
        ----------
        stream : writeable stream accepting bytes
           The stream into which the candidate information should be serialized.

        Raises
        ------
        May raise :exc:`toasty.pipeline.NotActionableError` if it turns out that this
        candidate is not one that can be imported into WWT.

        Returns
        -------
        None.

        """


# The PipelineManager class that orchestrates it all

IMAGE_SOURCE_CLASS_LOADERS = {}

def _load_astropix_image_source_class():
    from .astropix import AstroPixImageSource
    return AstroPixImageSource
IMAGE_SOURCE_CLASS_LOADERS['astropix'] = _load_astropix_image_source_class

def _load_djangoplicity_image_source_class():
    from .djangoplicity import DjangoplicityImageSource
    return DjangoplicityImageSource
IMAGE_SOURCE_CLASS_LOADERS['djangoplicity'] = _load_djangoplicity_image_source_class


class PipelineManager(object):
    _config = None
    _pipeio = None
    _workdir = None
    _img_source = None

    def __init__(self, workdir):
        self._workdir = workdir
        self._pipeio = PipelineIo.load_from_config(self._path('toasty-store-config.yaml'))

    def _path(self, *path):
        return os.path.join(self._workdir, *path)

    def _ensure_dir(self, *path):
        path = self._path(*path)
        os.makedirs(path, exist_ok=True)
        return path

    def ensure_config(self):
        if self._config is not None:
            return self._config

        self._ensure_dir()
        cfg_path = self._path('toasty-pipeline-config.yaml')

        if not os.path.exists(cfg_path):  # racey
            with open(cfg_path, 'wb') as f:
                self._pipeio.get_item('toasty-pipeline-config.yaml', dest=f)

        with open(cfg_path, 'rt', encoding='utf8') as f:
            config = yaml.safe_load(f)

        if config is None:
            raise Exception('no toasty-pipeline-config.yaml found in the storage')

        self._config = config
        return self._config

    def get_image_source(self):
        if self._img_source is not None:
            return self._img_source

        self.ensure_config()

        source_type = self._config.get('source_type')
        if not source_type:
            raise Exception('toasty pipeline configuration must have a source_type key')

        cls_loader = IMAGE_SOURCE_CLASS_LOADERS.get(source_type)
        if cls_loader is None:
            raise Exception('unrecognized image source type %s' % source_type)

        cls = cls_loader()
        cfg_key = cls.get_config_key()
        source_config = self._config.get(cfg_key)
        if source_config is None:
            raise Exception('no image source configuration key %s in the config file' % cfg_key)

        self._img_source = cls.deserialize(source_config)
        return self._img_source

    def process_todos(self):
        from ..builder import Builder
        from .. import par_util
        from ..pyramid import PyramidIO

        src = self.get_image_source()
        cand_dir = self._path('candidates')
        self._ensure_dir('cache_done')
        baseoutdir = self._ensure_dir('processed')

        # Lame hack to tidy up output slightly
        par_util.SHOW_INFORMATIONAL_MESSAGES = False

        for uniq_id in os.listdir(self._path('cache_todo')):
            cachedir = self._path('cache_todo', uniq_id)
            outdir = self._path('processed', uniq_id)

            pio = PyramidIO(outdir, scheme='LXY', default_format='png')
            builder = Builder(pio)
            cdata = open(os.path.join(cand_dir, uniq_id), 'rb')

            print(f'processing {uniq_id} ... ', end='')
            sys.stdout.flush()

            src.process(uniq_id, cdata, cachedir, builder)
            cdata.close()
            builder.write_index_rel_wtml()
            print('done')

            # Woohoo, done!
            os.rename(cachedir, self._path('cache_done', uniq_id))

    def publish(self):
        done_dir = self._ensure_dir('published')
        todo_dir = self._path('approved')
        pfx = todo_dir + os.path.sep

        for uniq_id in os.listdir(todo_dir):
            # If there's a index.wtml file, save it for last -- that will
            # indicate that this directory has uploaded fully successfully.

            filenames = os.listdir(os.path.join(todo_dir, uniq_id))

            try:
                index_index = filenames.index('index.wtml')
            except ValueError:
                pass
            else:
                temp = filenames[-1]
                filenames[-1] = 'index.wtml'
                filenames[index_index] = temp

            print(f'publishing {uniq_id} ...')

            for filename in filenames:
                # Get the components of the item path relative to todo_dir.
                sub_components = [todo_dir, uniq_id, filename]
                p = os.path.join(*sub_components)
                assert p.startswith(pfx)

                with open(p, 'rb') as f:
                    self._pipeio.put_item(*sub_components[1:], source=f)

            os.rename(os.path.join(todo_dir, uniq_id), os.path.join(done_dir, uniq_id))



# -*- mode: python; coding: utf-8 -*-
# Copyright 2019-2020 the AAS WorldWide Telescope project
# Licensed under the MIT License.

"""Generate tiles from a collection of images on a common TAN projection.

TODO: shuld be migrated to wwt_data_formats.

"""
from __future__ import absolute_import, division, print_function

__all__ = '''
MultiTanDataSource
'''.split()

import numpy as np

from .pyramid import Pos, next_highest_power_of_2

MATCH_HEADERS = [
    'CTYPE1', 'CTYPE2',
    'CRVAL1', 'CRVAL2',
]

# In my sample set, some of these files vary minutely from one header to the
# next, so we save them but don't demand exact matching. We should use
# np.isclose() or whatever it is.
SAVE_HEADERS = [
    'CD1_1', 'CD2_2',
]


class MultiTanDataSource(object):
    """Generate tiles from a collection of images on a common TAN projection.

    Some large astronomical images are stored as a collection of sub-images
    that share a common tangential projection, a format is that is nice and
    easy to convert into a WWT "study" tile pyramid. This class can process a
    collection of such images and break them into the highest-resolution layer
    of such a tile pyramid.

    """
    _ra_deg = None
    "The RA of the center of the TAN projection on the sky, in degrees."
    _dec_deg = None
    "The declination of the center of the TAN projection on the sky, in degrees."
    _rot_rad = None
    "The rotation of the TAN projection on the sky, in radians."
    _width = None
    "The width of the region in which image data are available, in pixels."
    _height = None
    "The height of the region in which image data are available, in pixels."
    _scale_x = None
    "The horizontal pixel scale, in degrees per pixel. May be negative."
    _scale_y = None
    "The vertical pixel scale, in degrees per pixel. Always positive."
    _tile_levels = None
    "How many levels there are in the tile pyramid for this study."
    _p2w = None
    """The width of the highest-resolution tiling, in pixels. This is the
    first power of 2 greater than or equal to _width."""
    _p2h = None
    """The height of the highest-resolution tiling, in pixels. This is the
    first power of 2 greater than or equal to _height."""
    _center_crx = None
    """The X position of the image center, in pixels relative to the center of
    the tangential projection."""
    _center_cry = None
    """The Y position of the image center, in pixels relative to the center of
    the tangential projection."""
    _crxmin = None
    "The minimum observed pixel X index, relative to the CRPIX of the projection."
    _crymin = None
    "The minimum observed pixel Y index, relative to the CRPIX of the projection."

    def __init__(self, paths, hdu_index=0):
        self._paths = paths
        self._hdu_index = hdu_index


    def _input_hdus(self):
        from astropy.io import fits

        for path in self._paths:
            with fits.open(path) as hdu_list:
                yield path, hdu_list[self._hdu_index]


    def compute_global_pixelization(self):
        """Read the input images to determine the global pixelation of this data set.

        This function reads the FITS headers of all of the input data files to
        determine the overall image size and the parameters of its tiling as a
        WWT study.

        """
        ref_headers = None
        crxmin = None

        for path, hdu in self._input_hdus():
            if ref_headers is None:
                ref_headers = {}

                for header in MATCH_HEADERS:
                    ref_headers[header] = hdu.header[header]

                for header in SAVE_HEADERS:
                    ref_headers[header] = hdu.header[header]

                if ref_headers['CTYPE1'] != 'RA---TAN':
                    raise Exception('all inputs must be in a TAN projection, but {} is not'.format(path))
                if ref_headers['CTYPE2'] != 'DEC--TAN':
                    raise Exception('all inputs must be in a TAN projection, but {} is not'.format(path))
            else:
                for header in MATCH_HEADERS:
                    expected = ref_headers[header]
                    observed = hdu.header[header]

                    if observed != expected:
                        raise Exception('inputs are not on uniform WCS grid; in file {}, expected '
                                        'value {} for header {} but observed {}'.format(path, expected, header, observed))

            crpix1 = hdu.header['CRPIX1'] - 1
            crpix2 = hdu.header['CRPIX2'] - 1

            this_crxmin = 0 - crpix1
            this_crxmax = (hdu.shape[1] - 1) - crpix1
            this_crymin = 0 - crpix2
            this_crymax = (hdu.shape[0] - 1) - crpix2

            if crxmin is None:
                crxmin = this_crxmin
                crxmax = this_crxmax
                crymin = this_crymin
                crymax = this_crymax
            else:
                crxmin = min(crxmin, this_crxmin)
                crxmax = max(crxmax, this_crxmax)
                crymin = min(crymin, this_crymin)
                crymax = max(crymax, this_crymax)

        # Figure out the global properties of the tiled TAN representation
        self._crxmin = crxmin
        self._crymin = crymin

        self._width = int(crxmax - self._crxmin) + 1
        self._height = int(crymax - self._crymin) + 1
        self._p2w = next_highest_power_of_2(self._width)
        self._p2h = next_highest_power_of_2(self._height)

        if self._p2w != self._p2h:  # TODO: figure this out and make it work.
            raise Exception('TODO: we don\'t properly handle non-square-ish images; got {},{}'.format(self._p2w, self._p2h))
        p2max = max(self._p2w, self._p2h)
        self._tile_levels = int(np.log2(p2max / 256))
        nfullx = self._p2w // 256
        nfully = self._p2h // 256

        self._ra_deg = ref_headers['CRVAL1']
        self._dec_deg = ref_headers['CRVAL2']

        cd1_1 = ref_headers['CD1_1']
        cd2_2 = ref_headers['CD2_2']
        cd1_2 = ref_headers.get('CD1_2', 0.0)
        cd2_1 = ref_headers.get('CD2_1', 0.0)

        if cd1_1 * cd2_2 - cd1_2 * cd2_1 < 0:
            cd_sign = -1
        else:
            cd_sign = 1

        self._rot_rad = np.arctan2(-cd_sign * cd1_2, cd2_2)
        self._scale_x = np.sqrt(cd1_1**2 + cd2_1**2) * cd_sign
        self._scale_y = np.sqrt(cd1_2**2 + cd2_2**2)

        self._center_crx = self._width // 2 + self._crxmin
        self._center_cry = self._height // 2 + self._crymin

        return self  # chaining convenience

    def create_wtml(
            self,
            name = 'MultiTan',
            url_prefix = './',
            fov_factor = 1.7,
            bandpass = 'Visible',
            description_text = '',
            credits_text = 'Created by toasty, part of the AAS WorldWide Telescope.',
            credits_url = '',
            thumbnail_url = '',
    ):
        """Create a WTML document with the proper metadata for this data set.

        :meth:`compute_global_pixelization` must have been called first.

        Parameters
        ----------
        name : str, defaults to "MultiTan"
          The dataset name to embed in the WTML file.
        url_prefix : str, default to "./"
          The beginning of the URL path to the tile data. The URL given in
          the WTML will be "URL_PREFIX{1}/{3}/{3}_{2}.png"
        fov_factor : float, defaults to 1.7
          The WTML file specifies the height of viewport that the client should
          zoom to when viewing this image. The value used is *fov_factor* times
          the height of the image.
        bandpass : str, defaults to "Visible"
          The bandpass of the image, as chosen from a menu of options
          supported by the WTML format: "Gamma", "HydrogenAlpha", "IR",
          "Microwave", "Radio", "Ultraviolet", "Visible", "VisibleNight",
          "XRay".
        description_text : str, defaults to ""
          Free text describing what this image is.
        credits_text : str, defaults to a toasty credit
          A brief textual credit of who created and processed the image data.
        credits_url: str, defaults to ""
          A URL with additional image credit information, or the empty string
          if no such URL is available.
        thumbnail_url : str, defaults to ""
          A URL of a thumbnail image (96x45 JPEG) representing this image data
          set, or the empty string for a default to be used.

        Returns
        -------
        folder : xml.etree.ElementTree.Element
          An XML element containing the WWT metadata.

        Examples
        --------
        To convert the returned XML structure into text, use
        :func:`xml.etree.ElementTree.tostring`:
        >>> from xml.etree import ElementTree as etree
        >>> folder = data_source.create_wtml()
        >>> print(etree.tostring(folder))

        """
        from xml.etree import ElementTree as etree

        folder = etree.Element('Folder')
        folder.set('Name', name)
        folder.set('Group', 'Explorer')

        place = etree.SubElement(folder, 'Place')
        place.set('Name', name)
        place.set('DataSetType', 'Sky')
        place.set('Rotation', str(self._rot_rad * 180 / np.pi))
        place.set('Angle', '0')
        place.set('Opacity', '100')
        place.set('RA', str(self._ra_deg / 15))  # this RA is in hours
        place.set('Dec', str(self._dec_deg))
        place.set('ZoomLevel', str(self._height * self._scale_y * fov_factor * 6))  # zoom = 6 * (FOV height in deg)
        # skipped: Constellation, Classification, Magnitude, AngularSize

        fgimgset = etree.SubElement(place, 'ForegroundImageSet')

        imgset = etree.SubElement(fgimgset, 'ImageSet')
        imgset.set('Name', name)
        imgset.set('Url', url_prefix + '{1}/{3}/{3}_{2}.png')
        imgset.set('WidthFactor', '2')
        imgset.set('BaseTileLevel', '0')
        imgset.set('TileLevels', str(self._tile_levels))
        imgset.set('BaseDegreesPerTile', str(self._scale_y * self._p2h))
        imgset.set('FileType', '.png')
        imgset.set('BottomsUp', 'False')
        imgset.set('Projection', 'Tan')
        imgset.set('CenterX', str(self._ra_deg))
        imgset.set('CenterY', str(self._dec_deg))
        imgset.set('OffsetX', str(self._center_crx * np.abs(self._scale_x)))
        imgset.set('OffsetY', str(self._center_cry * self._scale_y))
        imgset.set('Rotation', str(self._rot_rad * 180 / np.pi))
        imgset.set('DataSetType', 'Sky')
        imgset.set('BandPass', bandpass)
        imgset.set('Sparse', 'True')

        credits = etree.SubElement(imgset, 'Credits')
        credits.text = credits_text

        credurl = etree.SubElement(imgset, 'CreditsUrl')
        credurl.text = credits_url

        thumburl = etree.SubElement(imgset, 'ThumbnailUrl')
        thumburl.text = thumbnail_url

        desc = etree.SubElement(imgset, 'Description')
        desc.text = description_text

        return folder


    def generate_deepest_layer_numpy(
            self,
            pio,
            percentiles = [1, 99],
    ):
        """Fill in the deepest layer of the tile pyramid with Numpy-format data.

        Parameters
        ----------
        pio : :class:`toasty.pyramid.PyramidIO`
          A :class:`~toasty.pyramid.PyramidIO` instance to manage the I/O with
          the tiles in the tile pyramid.
        percentiles : iterable of numbers
          This is a list of percentile points to calculate while reading the
          data. Each number should be between 0 and 100. For each
          high-resolution tile, the percentiles are computed; then the *median*
          across all tiles is computed and returned.

        Returns
        -------
        percentiles : dict mapping numbers to numbers
          This dictionary contains the result of the median-percentile
          computation. The keys are the values provided in the *percentiles*
          parameter. The values are the median of each percentile across
          all of the tiles.

        Notes
        -----
        The implementation assumes that if individual images overlap, we can
        just use the pixels from any one of them without caring which
        particular one we choose.

        Because this operation involves reading the complete image data set,
        it offers a convenient opportunity to do some statistics on the data.
        This motivates the presence of the *percentiles* feature.

        """
        crxofs = (self._p2w - self._width) // 2
        cryofs = (self._p2h - self._height) // 2

        percentile_data = {p: [] for p in percentiles}

        for path, hdu in self._input_hdus():
            crpix1 = hdu.header['CRPIX1'] - 1
            crpix2 = hdu.header['CRPIX2'] - 1

            # (inclusive) image bounds in global pixel coords, which range
            # from 0 to p2{w,h} (non-inclusive), and have y=0 at the top. The FITS
            # data have y=0 at the bottom, so we need to flip them vertically.
            img_gx0 = int((crxofs - self._crxmin) - crpix1)
            img_gx1 = img_gx0 + hdu.shape[1] - 1
            img_gy1 = self._p2h - 1 - int((cryofs - self._crymin) - crpix2)
            img_gy0 = img_gy1 - (hdu.shape[0] - 1)

            assert img_gx0 >= 0
            assert img_gy0 >= 0
            assert img_gx1 < self._p2w
            assert img_gy1 < self._p2h

            # Tile indices at the highest resolution.

            ix0 = img_gx0 // 256
            iy0 = img_gy0 // 256
            ix1 = img_gx1 // 256
            iy1 = img_gy1 // 256

            # OK, load up the data, with a vertical flip, and grid them. While
            # we're doing that, calculate percentiles to inform the RGB-ification.

            data = hdu.data[::-1]
            n_tiles = 0

            pvals = np.nanpercentile(data, percentiles)
            for pct, pv in zip(percentiles, pvals):
                percentile_data[pct].append(pv)

            for iy in range(iy0, iy1 + 1):
                for ix in range(ix0, ix1 + 1):
                    # (inclusive) tile bounds in global pixel coords
                    tile_gx0 = ix * 256
                    tile_gy0 = iy * 256
                    tile_gx1 = tile_gx0 + 255
                    tile_gy1 = tile_gy0 + 255

                    # overlap (= intersection) of the image and the tile in global pixel coords
                    overlap_gx0 = max(tile_gx0, img_gx0)
                    overlap_gy0 = max(tile_gy0, img_gy0)
                    overlap_gx1 = min(tile_gx1, img_gx1)
                    overlap_gy1 = min(tile_gy1, img_gy1)

                    # coordinates of the overlap in image pixel coords
                    img_overlap_x0 = overlap_gx0 - img_gx0
                    img_overlap_x1 = overlap_gx1 - img_gx0
                    img_overlap_xslice = slice(img_overlap_x0, img_overlap_x1 + 1)
                    img_overlap_y0 = overlap_gy0 - img_gy0
                    img_overlap_y1 = overlap_gy1 - img_gy0
                    img_overlap_yslice = slice(img_overlap_y0, img_overlap_y1 + 1)

                    # coordinates of the overlap in this tile's coordinates
                    tile_overlap_x0 = overlap_gx0 - tile_gx0
                    tile_overlap_x1 = overlap_gx1 - tile_gx0
                    tile_overlap_xslice = slice(tile_overlap_x0, tile_overlap_x1 + 1)
                    tile_overlap_y0 = overlap_gy0 - tile_gy0
                    tile_overlap_y1 = overlap_gy1 - tile_gy0
                    tile_overlap_yslice = slice(tile_overlap_y0, tile_overlap_y1 + 1)

                    ###print(f'   {ix} {iy} -- {overlap_gx0} {overlap_gy0} {overlap_gx1} {overlap_gy1} -- '
                    ###      f'{tile_overlap_x0} {tile_overlap_y0} {tile_overlap_x1} {tile_overlap_y1} -- '
                    ###      f'{img_overlap_x0} {img_overlap_y0} {img_overlap_x1} {img_overlap_y1}')

                    pos = Pos(self._tile_levels, ix, iy)
                    p = pio.tile_path(pos)

                    try:
                        a = np.load(p)
                    except IOError as e:
                        if e.errno == 2:
                            a = np.empty((256, 256))
                            a.fill(np.nan)
                        else:
                            raise

                    a[tile_overlap_yslice,tile_overlap_xslice] = data[img_overlap_yslice,img_overlap_xslice]
                    np.save(p, a)

        percentile_data = {p: np.median(a) for p, a in percentile_data.items()}
        return percentile_data
